def tryfloat(x):
    try:
        return float(x)
    except ValueError:
        return float(0)

def read_radex(file,flow,fupp,bw=0.01,debug=False):
    """ 
    less hack-ey way to read radex.out files
    """
    linenum = 0
    line = file.readline()
    linenum+=1
    if debug: print line
    if line == '':
        return 0
    words = line.split()
    if words[1] == '--':
        freq = tryfloat(words[4])
    else:
        freq = 0
    while not(freq*(1-bw) < flow < freq*(1+bw)):
        if words[1] == 'T(kin)':
            tkin = tryfloat(words[3])
        elif line.find("Density of H2") != -1:
            dens = tryfloat(words[5])
        elif line.find("Density of pH2") != -1:
            pdens = tryfloat(words[5])
        elif line.find("Density of oH2") != -1:
            odens = tryfloat(words[5])
        elif line.find("Column density") != -1:
            col = tryfloat(words[4])
        line = file.readline(); linenum+=1
        words = line.split()
        if words[1] == '--':
            freq = tryfloat(words[4])
    TexLow   = tryfloat(words[6])
    TauLow   = tryfloat(words[7])
    TrotLow  = tryfloat(words[8])
    FluxLow  = tryfloat(words[11])
    line = file.readline(); linenum+=1
    words = line.split()
    if words[1] == '--':
        freq = tryfloat(words[4])
    while  not(freq*(1-bw) < fupp < freq*(1+bw)):
        line = file.readline(); linenum+=1
        if debug: print freq,flow,line
        words = line.split()
        if words[1] == '--':
            freq = tryfloat(words[4])
    TexUpp   = tryfloat(words[6])
    TauUpp   = tryfloat(words[7])
    TrotUpp  = tryfloat(words[8])
    FluxUpp  = tryfloat(words[11])
    while len(words) > 0 and words[1] == '--':
        line = file.readline(); linenum+=1
        words = line.split()
    return tkin,dens,col,TexLow,TexUpp,TauLow,TauUpp,TrotLow,TrotUpp,FluxLow,FluxUpp



from astropy import units as u
import sys
import os
import errno

def mkdir_p(path):
    """ mkdir -p equivalent [used by get_datafile]"""
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def united(qty, unit):
    if isinstance(qty,u.Quantity):
        return qty.to(unit)
    else:
        return qty*u.Unit(unit)

def uvalue(qty, unit):
    return united(qty, unit).value

def get_datafile(species, savedir='./'):
    """
    Load a molecular data file and save it into the specified directory
    """
    from astroquery import lamda

    datapath = os.path.join(savedir,species)

    species,suffix = os.path.splitext(species)
    if suffix == "":
        datapath += ".dat"
    elif suffix != ".dat":
        raise ValueError("Molecular data file must either be a species name or species.dat")

    if not os.path.isdir(savedir):
        mkdir_p(savedir)

    if not os.path.isfile(datapath):
        data = lamda.query(species, return_datafile=True)
        with open(datapath,'w') as out:
            out.writelines([d+"\n" for d in data])

    return os.path.split(datapath)

def verify_collisionratefile(fn):
    """
    Verify that a RADEX collisional rate file is valid to avoid a RADEX crash
    """
    from astroquery import lamda

    with open(fn,'r') as datafile:
        data = [s.strip() for s in datafile.readlines()]

        for qt in lamda.core.query_types:
            try:
                tbl = lamda.core._parse_datafile(data, qt)
            except Exception as ex:
                raise type(ex), type(ex)("Data file verification failed.  The molecular data file may be corrupt." +
                                         "\nOriginal Error in the parser: " +
                                         ex.args[0]), sys.exc_info()[2]
            if len(tbl) == 0:
                raise ValueError("No data found in the table for the category %s" % qt)



from .. import Radex
import numpy as np

def buildgrid(densities=np.logspace(0,8,20), abundance=10**-8.5, fortho=1e-3, **kwargs):

    tau = np.zeros_like(densities)

    R = Radex(species='co', abundance=abundance, collider_densities={'oH2':10,'pH2':10})
    for jj,dd in enumerate(densities):
        R.density = {'oH2':dd*fortho,'pH2':dd*(1-fortho)}
        R.abundance = abundance # reset column to the appropriate value
        R.run_radex(**kwargs)
        tau[jj] = R.tau[0]

    return tau

def try_variants():

    taud = {}
    for reuse_last in (True,False):
        for reload_molfile in (True,False):
            s = ('T' if reuse_last else 'F') + ('T' if reload_molfile else 'F')
            taud[s] = buildgrid(reuse_last=reuse_last, reload_molfile=reload_molfile)

    return taud



"""
Tools to generate synthetic spectra given a table of line strengths
"""
import numpy as np

from astropy import logging
from astropy.modeling import models
from astropy import units as u
from astropy import constants as c

try:
    import specutils
    Spectrum = specutils.Spectrum1D
    if not hasattr(specutils.wcs, 'Spectrum1DLookupWCS'):
        raise ImportError("Wrong version of specutils.")
    Spectrum1DLookupWCS = specutils.wcs.Spectrum1DLookupWCS

    class FrequencyArray(Spectrum1DLookupWCS):
        """
        Generate a frequency array given a minimum frequency, maximum frequency,
        and # of points

        Parameters
        ----------
        minfreq: u.Quantity
        maxfreq: u.Quantity
            Minimum and maximum frequency to plot
        npts: int
            The number of spectral points to include
        """
        def __init__(self, minfreq, maxfreq, npts):
            self.minfreq = minfreq
            self.maxfreq = maxfreq
            self.npts = npts
            self.delta = (maxfreq-minfreq)/npts
            self.unit = u.Hz
            self.refpix = 0
            self.refval = minfreq
            #super(FrequencyArray, self).__init__(minfreq, (maxfreq-minfreq)/npts, 0, u.Hz)

        def __call__(self, pixel_indices):
            return (self.refval + self.delta * (pixel_indices -
                                                self.refpix))

        @property
        def lookuptable(self):
            if not hasattr(self, '_lookuptable'):
                self._lookuptable = self(np.arange(npts))
            return self._lookuptable


    class SyntheticSpectrum(Spectrum):
        """
        Synthetic Spectrum class - neato!
        """

        def __init__(self, wcs, species, linewidth):
            self.wcs = wcs
            self.species = species
            self.linewidth = linewidth

        @classmethod
        def from_table(cls, wcs, table, species,
                       linewidth=1.0*u.km/u.s,
                       profile_function=models.Gaussian1D):
            """
            Create a synthetic spectrum from a RADEX (or DESPOTIC, eventually)
            output

            Parameters
            ----------
            wcs: SpectralWCS
                A spectral world coordinate system.  You can generate one with
                FrequencyArray or specutils.wcs.Spectrum1DLookupWCS
            table: astropy.Table
                Result of the RADEX query (from R.get_table())
            linewidth: u.Quantity (km/s)
                The width of the line to plot
            npts: int
                The number of spectral points to include
            profile_function: astropy.modeling.model
                The model function to use.  Must accept, in order:
                
                 * flux (peak)
                 * frequency center (Hz)
                 * frequency width (Hz)
                 

            Examples
            --------
            >>> from pyradex import Radex,synthspec
            >>> R = Radex(species='ch3cn')
            >>> R.run_radex()
            >>> wcs = synthspec.FrequencyArray(91.95*u.GHz, 92*u.GHz, npts=1000)
            >>> S = synthspec.SyntheticSpectrum.from_table(wcs, R.get_Table(),
            ...                                 species='ch3cn')
            >>> S.plot()
            """

            self = cls(wcs, species, linewidth)

            self.profile_function = profile_function

            self.species = species
            self.linewidth = linewidth
            self.wcs = wcs
            if hasattr(wcs,'minfreq'):
                self.minfreq,self.maxfreq = wcs.minfreq,wcs.maxfreq
            else:
                self.minfreq,self.maxfreq = wcs.min(),wcs.max()

            linefreqs = u.Quantity(table['frequency'],
                                   unit=u.Unit(table['frequency'].unit))
            self.table = table[(linefreqs>self.minfreq) & (linefreqs<self.maxfreq)]
            self.linefreqs = linefreqs
            self.width_frequency = (linewidth/c.c *
                                    u.Quantity(self.table['frequency'],
                                               unit=u.Unit(self.table['frequency'].unit)))
            self.T_B = self.table['T_B']

            data = self.get_profile()

            super(Spectrum,self).__init__(data=data, wcs=self.wcs,
                                          unit=u.Unit(table['T_B'].unit))

            return self

        @classmethod
        def from_RADEX(cls, wcs, rad,
                       linewidth=1.0*u.km/u.s,
                       profile_function=models.Gaussian1D):
            """
            Create a synthetic spectrum from a RADEX class

            Parameters
            ----------
            wcs: SpectralWCS
                A spectral world coordinate system.  You can generate one with
                FrequencyArray or specutils.wcs.Spectrum1DLookupWCS
            rad: pyradex.Radex instance
                Result of the RADEX query
            linewidth: u.Quantity (km/s)
                The width of the line to plot
            npts: int
                The number of spectral points to include
            profile_function: astropy.modeling.model
                The model function to use.  Must accept, in order:
                
                 * flux (peak)
                 * frequency center (Hz)
                 * frequency width (Hz)
                 

            Examples
            --------
            >>> from pyradex import Radex,synthspec
            >>> R = Radex(species='ch3cn')
            >>> R.run_radex()
            >>> wcs = synthspec.FrequencyArray(91.95*u.GHz, 92*u.GHz, npts=1000)
            >>> S = synthspec.SyntheticSpectrum.from_RADEX(wcs, R)
            >>> S.plot()
            """

            self = cls(wcs, rad.species, linewidth)

            self.profile_function = profile_function

            self.wcs = wcs
            if hasattr(wcs,'minfreq'):
                self.minfreq,self.maxfreq = wcs.minfreq,wcs.maxfreq
            else:
                self.minfreq,self.maxfreq = wcs.min(),wcs.max()

            self.rad = rad
            linefreqs = rad.frequency
            linefreq_mask = (linefreqs>self.minfreq) & (linefreqs<self.maxfreq)
            included_frequencies_mask = linefreq_mask[rad.inds_frequencies_included]
            self.linefreqs = linefreqs[linefreq_mask]
            self.T_B = rad.T_B[included_frequencies_mask]
            self.width_frequency = (linewidth/c.c * self.linefreqs)

            data = self.get_profile()

            super(Spectrum,self).__init__(data=data, wcs=self.wcs,
                                          unit=u.Unit(rad.T_B.unit))

            return self

        def get_profile(self, velocity_offset=0*u.km/u.s):
            
            def model(xpts):
                if isinstance(xpts,u.Quantity):
                    xpts = xpts.to(u.Hz).value
                M = np.zeros_like(xpts)
                freqs = self.linefreqs + (self.linefreqs*velocity_offset/c.c)
                for freq,flux,width in zip(freqs,
                                           self.T_B,
                                           self.width_frequency):
                    M += self.profile_function(flux.value, freq.to(u.Hz).value,
                                               width.to(u.Hz).value)(xpts)

                return M

            X = self.wcs(np.arange(self.wcs.npts))

            return model(X)

        def plot(self, update_data=False, *args, **kwargs):
            import pylab as pl

            if update_data:
                self.data = self.get_profile()

            pl.gca().set_xlabel(self.dispersion.unit.to_string())
            if hasattr(self.data,'unit'):
                pl.gca().set_ylabel(self.data.unit.to_string())
                data = self.data.value
            else:
                data = self.data

            return pl.plot(self.dispersion.value, data, *args, **kwargs)

        def __call__(self, linewidth=None, velocity_offset=0*u.km/u.s, **kwargs):
            """
            Return a synthetic spectrum created by calling RADEX.  Parameters
            are passed to pyradex.Radex (except linewidth)

            Parameters
            ----------
            linewidth: u.Quantity (km/s)
                The width of the line to plot

            Examples
            --------
            >>> from pyradex import Radex,synthspec
            >>> radex_pars = dict(temperature=20, column=1e13,
            ...                   abundance=10**-8.5,
            ...                   collider_densities={'H2':1e4})
            >>> R = Radex(species='oh2co-h2', **radex_pars)
            >>> R.run_radex()
            >>> wcs = synthspec.FrequencyArray(4.828*u.GHz, 4.830*u.GHz, npts=1000)
            >>> S = synthspec.SyntheticSpectrum.from_RADEX(wcs, R)
            >>> S.plot()
            >>> radex_pars['temperature'] = 50
            >>> S2 = S(velocity_offset=2*u.km/u.s, **radex_pars)
            >>> S2.plot()
            """

            from .core import Radex

            rad = Radex(species=self.species, **kwargs)

            if linewidth is None:
                linewidth = self.linewidth
            else:
                self.linewidth = linewidth

            self.rad = rad
            linefreqs = rad.frequency
            linefreq_mask = (linefreqs>self.minfreq) & (linefreqs<self.maxfreq)
            self.linefreqs = linefreqs[linefreq_mask]
            self.T_B = rad.T_B[linefreq_mask]
            self.width_frequency = (linewidth/c.c * self.linefreqs)

            self.data = self.get_profile()

            return self
except ImportError as exc:
    logging.warn("Failed to import synthetic spectrum package: "+exc.args[0])



from __future__ import print_function
import tarfile
import re
import os
import shutil
try:
    from astropy.utils.data import download_file
except ImportError:
    download_file = None


def install_radex(download=True, extract=True, patch=True, compile=True):
    if download:
        filename = download_radex()
    if extract:
        extract_radex(filename)
    if patch:
        patch_radex()
    if compile:
        compile_radex()

def download_radex(redownload=True,
                   url='http://www.sron.rug.nl/~vdtak/radex/radex_public.tar.gz'):

    filename = 'radex_public.tar.gz'

    if os.path.isfile(filename) and not redownload:
        return filename

    print("Downloading RADEX")

    try:
        filename = download_file(url, cache=True)
    except:
        import requests
        r = requests.get(url,
                         #data={'filename':filename},
                         stream=True,
                         verify=False)
        with open(filename,'wb') as f:
            for chunk in r.iter_content(1024):
                f.write(chunk)

    print("Download succeeded, or at least didn't obviously fail.")

    return filename

def extract_radex(filename='radex_public.tar.gz'):
    with tarfile.open(filename,mode='r:gz') as tar:
        # this will fail if readonly files are overwritten: tf.extractall()
        for file_ in tar:
            try:
                tar.extract(file_)
            except IOError as e:
                os.remove(file_.name)
                tar.extract(file_)
            finally:
                os.chmod(file_.name, file_.mode)

def patch_radex():
    # PATCHES
    radlines = []
    vers = re.compile("parameter\(version = '([a-z0-9]*)'\)")
    with open('Radex/src/radex.inc') as f:
        for line in f.readlines():
            if ('parameter(radat' in line or
                'parameter(version' in line or
                'parameter(logfile' in line or
                'parameter (method' in line):
                line = 'c'+line
            if vers.search(line):
                radlines.append("c      version = '%s'\n" % vers.search(line).groups()[0])
            if 'parameter(debug' in line:
                line = 'c'+line
                radlines.append('      common/dbg/debug\n')
            radlines.append(line)

            if 'parameter (method = 3)' in line:
                radlines.append('      common/setup/radat,method,version,logfile\n')

    with open('Radex/src/radex.inc','w') as f:
        for line in radlines:
            f.write(line)

    with open('Radex/src/background.f') as f:
        lines = f.readlines()

    with open('Radex/src/background.f','w') as f:
        for line in lines:
            if 'parameter(huge' in line:
                f.write('       parameter(huge=1.0e38)\n')
                f.write('c      ! highest allowed by f90 (fdvt 28apr06)\n')
            else:
                f.write(line)

"""
Works for hpc:
    PATH=/Users/adam/repos/hpc/bin/:/usr/bin:~/virtual-python/bin/:/bin FFLAGS='-m64 -fPIC' CFLAGS='-fno-strict-aliasing -fno-common -dynamic -m64 -g -O2' LDFLAGS='-m64 -undefined dynamic_lookup -bundle' python -c "import install_radex; install_radex.compile_radex(f77exec='/Users/adam/repos/hpc/bin/gfortran')"

Works for 4.2.3:
    FFLAGS='-arch i686 -arch x86_64 -fPIC' CFLAGS='-fno-strict-aliasing -fno-common -dynamic -arch i386 -arch x86_64 -g -O2' LDFLAGS='-arch i686 -arch x86_64 -undefined dynamic_lookup -bundle' python setup.py install_radex
"""

def compile_radex(fcompiler='gfortran',f77exec=None):
    r1 = os.system('f2py -h pyradex/radex/radex.pyf Radex/src/*.f --overwrite-signature > radex_build.log')
    if r1 != 0:
        raise SystemError("f2py failed with error %i" % r1)

    if f77exec is None:
        f77exec=''
    else:
        f77exec = '--f77exec=%s' % f77exec
    cmd = 'f2py -m radex -c Radex/src/*.f --fcompiler=%s %s >> radex_build.log' % (fcompiler,f77exec)
    r2 = os.system(cmd)
    if r2 != 0:
        raise SystemError("f2py failed with error %i" % r2)

    r3 = os.system('mv radex.so pyradex/radex/')
    if r3 != 0:
        raise SystemError("moving failed with error %i; radex.so was not created successfully" % r3)

def build_radex_executable(datapath='./'):
    filename = download_radex(redownload=False)
    # need to re-extract the RADEX source to get an un-patched version
    extract_radex(filename)
    compile_radex_source(datapath=datapath)


def compile_radex_source(datapath='./'):
    """
    Compile the source file in the Radex/ directory

    May be good to download & untar first
    """

    cwd = os.getcwd()

    # Set datapath to the 'examples' directory if not specified (since there is
    # at least co.dat there)
    if datapath is None:
        datapath = os.path.join(os.getcwd(), 'examples')
        print("Datapath was not defined.  Set to ",datapath)

    # make sure ~ gets expanded
    datapath = os.path.expanduser(datapath)

    try:
        os.link('Radex/data/hco+.dat',os.path.join(datapath,'hco+.dat'))
    except OSError:
        pass

    # I think fortran requires a trailing slash... can't hurt
    if datapath[-1] != '/':
        datapath = datapath+'/'

    os.chdir('Radex/src/')
    with open('radex.inc','r') as f:
        lines = [L.replace('/Users/floris/Radex/moldat/',datapath)
                 if 'radat' in L else L
                 for L in f.readlines()]
    with open('radex.inc','w') as of:
        of.writelines(lines)

    method_types = {1:'sphere',2:'lvg',3:'slab'}
    for method in method_types:
        radex_inc_method('./',method=method)
        r1 = os.system('make')
        if r1 != 0:
            raise SystemError("radex make failed with error %i" % r1)
        shutil.move('../bin/radex','../bin/radex_%s' % method_types[method])

    os.chdir(cwd)

def radex_inc_method(datapath, method=1):
    """
    Convert the radex.inc file to a method

    Parameters
    ----------
    datapath: path
        a directory path containing the target radex.inc
    method: 1,2,3
        1: sphere
        2: lvg
        3: slab
    """
    fn = os.path.join(datapath, 'radex.inc')
    with open(fn,'r') as f:
        lines = [L.replace('/Users/floris/Radex/moldat/',datapath)
                 if 'radat' in L else L
                 for L in f.readlines()]
    
    radlines = []
    for line in lines:
        if ('parameter (method' in line):
            line = 'c'+line
        if 'method = 3' in line:
            radlines.append('      parameter (method = %i)\n' % method)
        radlines.append(line)

    with open(fn,'w') as f:
        f.writelines(radlines)




try:
    import despotic
except ImportError:
    despotic=False
import numpy as np
import warnings
import os
from collections import defaultdict
from .utils import united,uvalue

try:
    from astropy import units as u
    from astropy import constants
    from astropy.table import Table,Column
except ImportError:
    u = False

class Despotic(object):
    """
    A class meant to be similar to RADEX in terms of how it is called but to
    use Despotic instead of RADEX as the backend
    """

    # escapeprobabilty geometry names
    _epgdict = {'lvg':'LVG','sphere':'sphere','slab':'slab'}

    def __call__(self, **kwargs):
        # reset the parameters appropriately
        self.__init__(**kwargs)
        return self.lineLum()

    def __init__(self,
                 collider_densities={'ph2':990,'oh2':10},
                 temperature=30,
                 species='co',
                 datapath=None,
                 hcolumn=1e21,
                 abundance=1e-5,
                 #column=1e13,
                 tbackground=2.7315,
                 deltav=1.0,
                 escapeProbGeom='lvg',
                 outfile='radex.out',
                 logfile='radex.log',
                 debug=False,
                 ):
        """
        Interface to DESPOTIC

        Parameters
        ----------
        collider_densities: dict
            Dictionary giving the volume densities of the collider(s) in units of
            cm^-3.  Valid entries are h2,oh2,ph2,e,He,H,H+.  The keys are
            case-insensitive.
        temperature: float
            Local gas temperature in K
        species: str
            A string specifying a valid chemical species.  This is used to look
            up the specified molecule
        hcolumn: float
            The total column density of hydrogen.
        abundance: float
            The molecule's abundance relative to H (NOT H2 as is normally done!).
        tbackground: float
            Background radiation temperature (e.g., CMB)
        deltav: float
            The FWHM line width (really, the single-zone velocity width to
            scale the column density by: this is most sensibly interpreted as a
            velocity gradient (dv/dR))
        sigmaNT: float
            Nonthermal velocity dispersion
            (this is strictly ignored - deltav IS sigmant)
        datapath: str
            Path to the molecular data files
        """

        if not despotic:
            raise ImportError("Despotic could not be imported.  Please check that it is installed.")
        
        self.cloud = despotic.cloud()
        self.cloud.nH = float(np.sum([collider_densities[k]*2 if 'h2' in k.lower()
                                      else collider_densities[k]
                                      for k in collider_densities]))

        for k in collider_densities.keys():
            collider_densities[k.lower()] = collider_densities[k]

        if 'ph2' in collider_densities:
            self.cloud.comp.xpH2 = collider_densities['ph2'] / self.cloud.nH
        if 'oh2' in collider_densities:
            self.cloud.comp.xoH2 = collider_densities['oh2'] / self.cloud.nH

        self.cloud.Td = uvalue(temperature,u.K)
        self.cloud.Tg = uvalue(temperature,u.K)
        self.cloud.dust.sigma10 = 0.0

        self.cloud.colDen = uvalue(hcolumn,u.cm**-2)


        if uvalue(tbackground,u.K) > 2.7315:
            self.cloud.rad.TradDust = uvalue(tbackground,u.K)

        self.species = species
        if datapath is None:
            emitterFile = species+'.dat'
        else:
            emitterFile = os.path.expanduser(os.path.join(datapath, species+'.dat'))
        self.cloud.addEmitter(species, abundance, emitterFile=emitterFile)

        self.cloud.comp.computeDerived(self.cloud.nH)

        self.escapeProbGeom = escapeProbGeom

        self.deltav = deltav


    def recompute(self):
        self.cloud.comp.computeDerived(self.cloud.nH)
        self.lineLum()

    @property
    def deltav(self):
        if self.cloud.sigmaNT > 0:
            return (self.cloud.sigmaNT*u.cm/u.s).to(u.km/u.s)
        elif self.cloud.dVdr > 0:
            return (self.cloud.dVdr*u.s**-1).to(u.km/u.s/u.pc)
        else:
            raise ValueError("The velocity gradient is zero")

    @deltav.setter
    def deltav(self, deltav):

        if self.escapeProbGeom == 'LVG':
            self._dv = united(self.cloud.dVdr,u.s**-1)
            # See notes.rst: DESPOTIC must have a different dVdR to get the
            # same results as RADEX
            # 1.0645 / sqrt(8*log(2)) = sqrt(2 * pi) / (8*log(2))
            self.cloud.dVdr = (uvalue(united(deltav,u.km/u.s/u.pc),u.s**-1) *
                               np.sqrt(8*np.log(2))*2)
            # (2*np.pi)**0.5/(8*np.log(2))

        else:
            self._dv = united(self.cloud.sigmaNT,u.km/u.s)

            FWHM = united(deltav,u.km/u.s)
            self.sigmaTot = FWHM/np.sqrt(8.0*np.log(2))
            self.cloud.sigmaNT = uvalue(np.sqrt(self.sigmaTot**2 -
                                                self.cs**2/self.cloud.emitters[self.species].data.molWgt),
                                        u.km/u.s)

    @property
    def cs(self):
        return np.sqrt(constants.k_B*united(self.cloud.Tg,u.K)/(self.cloud.comp.mu*constants.m_p)).to(u.km/u.s)

    def lineLum(self, **kwargs):
        if 'escapeProbGeom' not in kwargs:
            kwargs['escapeProbGeom'] = self.escapeProbGeom
        return self.cloud.lineLum(self.species, **kwargs)

    @property
    def escapeProbGeom(self):
        return self._epg

    @escapeProbGeom.setter
    def escapeProbGeom(self, escapeProbGeom):
        mdict = self._epgdict
        if escapeProbGeom.lower() not in mdict:
            raise ValueError("Invalid escapeProbGeom, must be one of "+",".join(mdict.values()))
        self._epg = mdict[escapeProbGeom]

    @property
    def density(self):
        d = {'H2':self.cloud.comp.xH2*self.nH,
             'oH2':self.cloud.comp.xoH2*self.nH,
             'pH2':self.cloud.comp.xpH2*self.nH,
             'e':self.cloud.comp.xe*self.nH,
             'H':self.cloud.comp.xHI*self.nH,
             'He':self.cloud.comp.xHe*self.nH,
             'H+':self.cloud.comp.xHplus*self.nH}
        if d['H2'] != 0 and (d['oH2'] != 0 or d['pH2'] != 0):
            # either ortho + para or total H2, not both
            d['H2'] = 0
        if u:
            for k in d:
                d[k] = d[k] * u.cm**-3
        return d

    @property
    def nH(self):
        return self.cloud.nH

    @nH.setter
    def nH(self, nh):
        self.cloud.nH = nh

    @property
    def nH2(self):
        return self.cloud.nH/2.

    @nH2.setter
    def nH2(self, nh2):
        self.nh = nh2*2.

    @property
    def beta(self):
        emitter = self.cloud.emitters[self.species]
        ep = [emitter.escapeProb[upper,lower]
              for (upper,lower) in zip(emitter.data.radUpper,
                                       emitter.data.radLower)]
        return ep

    @density.setter
    def density(self, collider_density):
        collider_densities = defaultdict(lambda: 0)
        for k in collider_density:
            collider_densities[k.upper()] = collider_density[k]

        if 'OH2' in collider_densities:
            if not 'PH2' in collider_densities:
                raise ValueError("If o-H2 density is specified, p-H2 must also be.")
            collider_densities['H2'] = (collider_densities['OH2'] + collider_densities['PH2'])
        elif 'H2' in collider_densities:
            warnings.warn("Using a default ortho-to-para ratio (which "
                          "will only affect species for which independent "
                          "ortho & para collision rates are given)")

            T = self.temperature.value if hasattr(self.temperature,'value') else self.temperature
            if T > 0:
                opr = min(3.0,9.0*np.exp(-170.6/T))
            else:
                opr = 3.0
            fortho = opr/(1+opr)
            collider_densities['OH2'] = collider_densities['H2']*fortho
            collider_densities['PH2'] = collider_densities['H2']*(1-fortho)

        total_density = np.sum([collider_densities[x] * (2 if '2' in x else 1)
                                for x in (['OH2','PH2','H','E','HE','H+'])])
        self.nH = total_density

        self.cloud.comp.xH2 = collider_densities['H2']/self.nH
        self.cloud.comp.xoH2 = (collider_densities['OH2'])/self.nH
        self.cloud.comp.xpH2 = (collider_densities['PH2'])/self.nH

        self.cloud.comp.xe = collider_densities['E']/self.nH
        self.cloud.comp.xHI = collider_densities['H']/self.nH
        self.cloud.comp.xHe = collider_densities['HE']/self.nH
        self.cloud.comp.xHplus = collider_densities['H+']/self.nH

        self.cloud.comp.computeDerived()

    @property
    def temperature(self):
        return self.cloud.Tg

    @property
    def upperlevelpop(self):
        return self.cloud.emitters[self.species].levPop[1:]

    @property
    def lowerlevelpop(self):
        return self.cloud.emitters[self.species].levPop[:-1]

    def get_table(self, **kwargs):

        D = self.lineLum(**kwargs)

        # remap names to match RADEX
        name_mapping = {'upper':'upperlevel',
                        'lower':'lowerlevel',
                        'freq':'frequency',}


        names = D[0].keys()
        T = Table(names=[name_mapping[n]
                         if n in name_mapping
                         else n
                         for n in names],
                  dtype=[type(D[0][k]) for k in names])
    
        for row in D:
            T.add_row([row[k] for k in names])

        T.add_column(Column(name='upperlevelpop',
                            data=self.upperlevelpop,
                            dtype='float'))
        T.add_column(Column(name='lowerlevelpop',
                            data=self.lowerlevelpop,
                            dtype='float'))

        return T



from __future__ import print_function
import subprocess
import tempfile
import numpy as np
import warnings
import astropy.units as u
from collections import defaultdict
import itertools
import os

from . import utils
from . import synthspec

try:
    from astropy import units as u
    from astropy import constants
    import astropy.table
except ImportError:
    u = False

__all__ = ['pyradex', 'write_input', 'parse_outfile', 'call_radex', 'Radex', 'density_distribution']


# silly tool needed for fortran misrepresentation of strings
# http://stackoverflow.com/questions/434287/what-is-the-most-pythonic-way-to-iterate-over-a-list-in-chunks
def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return itertools.izip_longest(*args, fillvalue=fillvalue)
        

def pyradex(executable='radex', minfreq=100, maxfreq=130,
            collider_densities={'H2':1}, debug=False, delete_tempfile=True,
            **kwargs):
    """
    Get the radex results for a set of input parameters


    Parameters
    ----------
    executable : str
        Full path to the RADEX executable
    minfreq : float
        Lowest frequency line to store, in GHz
        (note: any astropy.unit spectroscopic unit is also allowed)
    maxfreq : float
        Highest frequency line to store
    collider_densities : dict
        Collider names and their number densities
        If the molecule specified has both o-H2 and p-H2, you will get a
        WARNING if you specify 'H2'
        An ortho/para example:
        collider_densities = {'oH2':900, 'pH2':100} 
        which will yield H2 = 1000

    See write_input for additional parameters

    Returns
    -------
    An astropy table containing the RADEX returns

    .. WARNING:: If RADEX spits out *******, it will be replaced with -999
    """
    warnings.warn("pyradex is deprecated: it uses very slow hard disk file i/o.  Use pyradex.Radex instead if you can.")

    infile,outfile = write_input(minfreq=minfreq, maxfreq=maxfreq,
            delete_tempfile=delete_tempfile,
            collider_densities=collider_densities, **kwargs)

    logfile = call_radex(executable, infile.name, debug=debug,
                         delete_tempfile=delete_tempfile)

    check_logfile(logfile.name)

    data = parse_outfile(outfile.name)

    if debug:
        with open(infile.name,'r') as inf:
            print("Input:")
            print(inf.read())
        with open(outfile.name,'r') as out:
            print("Output:")
            print(out.read())

    infile.close()
    outfile.close()
    logfile.close()

    return data

def check_logfile(logfilename):
    with open(logfilename,'r') as f:
        if "Warning: Assuming thermal o/p ratio" in f.read():
            warnings.warn("Assumed thermal o/p ratio since only H2 was given but collider file has o- and p- H2")

def write_input(temperature=10, column=1e12, collider_densities={'H2':1},
        bw=0.01, tbg=2.73, species='co', velocity_gradient=1.0, minfreq=1,
        maxfreq=10, delete_tempfile=True):
    """
    Write radex.inp file parameters

    Parameters
    ----------
    temperature : float
        Kinetic temperature (K)
    collider_densities : dict
        Collider names and their number densities
    column : float
        column density of the molecule
    species : str
        Name of the molecule (specifically, the prefix for the file name, e.g.
        for "co.dat", species='co').  Case sensitive!
    tbg : float
        Temperature of the background radiation (e.g. CMB)
    velocity_gradient : float
        Velocity gradient per pc in km/s
    """

    if hasattr(minfreq, 'unit'):
        minfreq = minfreq.to('GHz',u.spectral()).value
    if hasattr(maxfreq, 'unit'):
        maxfreq = maxfreq.to('GHz',u.spectral()).value

    infile = tempfile.NamedTemporaryFile(mode='w', delete=delete_tempfile)
    outfile = tempfile.NamedTemporaryFile(mode='w', delete=delete_tempfile)
    infile.write(species+'.dat\n')
    infile.write(outfile.name+'\n')
    infile.write(str(minfreq)+' '+str(maxfreq)+'\n')
    infile.write(str(temperature)+'\n')

    # RADEX doesn't allow densities < 1e-3
    for k in collider_densities.keys():
        if collider_densities[k] < 1e-3:
            collider_densities.pop(k)

    infile.write('%s\n' % len(collider_densities))
    for name,dens in collider_densities.iteritems():
        infile.write('%s\n' % name)
        infile.write(str(dens)+'\n')
    infile.write(str(tbg)+'\n')
    infile.write(str(column)+'\n')
    infile.write(str(velocity_gradient)+'\n')
    # end the input file
    infile.write('0\n')
    infile.flush()
    return infile,outfile

def call_radex(executable, inpfilename, debug=False, delete_tempfile=True):

    logfile = tempfile.NamedTemporaryFile(mode='w', delete=delete_tempfile)
    cmd = '{radex} < {inpfile} > {logfile}'.format(
        radex=executable,
        inpfile=inpfilename,
        logfile=logfile.name)
    if debug:
        print("Command:",cmd)

    result = subprocess.call(cmd, shell=True)
    if result != 0:
        print("RADEX returned error code %i" % result)
        with open(logfile.name,'r') as f:
            print(f.read())

    return logfile


header_names = ['J_up','J_low','E_UP','FREQ', 'WAVE', 'T_EX', 'TAU', 'T_R', 'POP_UP', 'POP_LOW', 'FLUX_Kkms',   'FLUX_Inu']
header_units = [None,       None, u.K,   u.GHz,  u.um,   u.K,    None,  u.K,   None,     None,     u.K*u.km/u.s, u.erg/u.cm**2/u.s]
dtypes       = [str,   str,    float, float,  float,  float,  float, float,  float,    float,     float,         float]

def parse_outfile(filename):
    with open(filename,'r') as f:
        alllines = f.readlines()
        header = {L.split(":")[0][2:].strip():L.split(":")[1].strip()
                for L in alllines
                if L[0]=='*'}
        lines = [L.replace("--","  ") for L in alllines 
                if (L[0] != '*' 
                    and 'iterat' not in L 
                    and 'GHz' not in L 
                    and 'TAU' not in L)]
    data_list = [[x if '*' not in x else '-999' for x in L.split()] for L in lines]
    if len(data_list) == 0:
        raise ValueError("No lines included?")
    data_in_columns = map(list,zip(*data_list))
    columns = [astropy.table.Column(data=C, name=name.lower(), unit=unit, dtype=dtype) 
            for C,name,unit,dtype in zip(data_in_columns, header_names, header_units, dtypes)]
    data = astropy.table.Table(columns, meta=header)
    return data

class Radex(object):

    def __call__(self, return_table=True, **kwargs):
        # reset the parameters appropriately
        self.__init__(**kwargs)
        niter = self.run_radex()

        if return_table:
            return self.get_table()
        else:
            return niter

    def __init__(self,
                 collider_densities={'ph2':990,'oh2':10},
                 temperature=30,
                 species='co',
                 column=None,
                 h2column=None,
                 tbackground=2.7315,
                 deltav=1.0,
                 abundance=None,
                 datapath=None,
                 escapeProbGeom='lvg',
                 outfile='radex.out',
                 logfile='radex.log',
                 debug=False,
                 mu=2.8,
                 source_area=None,
                 ):
        """
        Direct wrapper of the radex FORTRAN code

        Parameters
        ----------
        collider_densities: dict
            Dictionary giving the volume densities of the collider(s) in units of
            cm^-3.  Valid entries are h2,oh2,ph2,e,He,H,H+.  The keys are
            case-insensitive.
        temperature: float
            Local gas temperature in K
        species: str
            A string specifying a valid chemical species.  This is used to look
            up the specified molecule
        column: float
            The column density of the molecule of interest.  If the column is
            specified, the abundance is equal to (column/(dv*total
            density*length)).
        h2column: float
            The column of h2. 
        abundance: float
            The molecule's abundance relative to the total collider density in
            each velocity bin, i.e. column = abundance * density * length * dv.
            If both abundance and column are specified, abundance is ignored.
        tbackground: float
            Background radiation temperature (e.g., CMB)
        deltav: float
            The FWHM line width (really, the single-zone velocity width to
            scale the column density by: this is most sensibly interpreted as a
            velocity gradient (dv/length))
        datapath: str
            Path to the molecular data files.  If it is not specified, defaults
            to the current directory, OR the shell variable RADEX_DATAPATH if
            it is specified.
        outfile: str
            Output file name
        logfile: str
            Log file name
        escapeProbGeom: 'lvg','sphere','slab'
            Which escape probability method to use
        mu: float
            Mean mass per particle in AMU.  Set to 2.8 for H2+Helium mix
        source_area: float / unit
            The emitting area of the source on the sky in steradians
        """
        from pyradex.radex import radex
        self.radex = radex

        self.mu = mu

        if os.getenv('RADEX_DATAPATH') and datapath is None:
            datapath = os.getenv('RADEX_DATAPATH')

        if datapath is not None:
            self.datapath = datapath
            if self.datapath != os.path.expanduser(datapath):
                raise ValueError("Data path %s was not successfully stored; instead %s was." % (datapath,self.datapath))
        self.molpath = os.path.join(self.datapath,species+'.dat')
        if self.molpath == '':
            raise ValueError("Must set a species name.")
        if not os.path.exists(self.molpath):
            raise ValueError("Must specify a valid path to a molecular data file else RADEX will crash."
                             "  Current path is {0}".format(self.molpath))

        self.density = collider_densities
        self.species = species

        self.outfile = outfile
        self.logfile = logfile
        self.escapeProbGeom = escapeProbGeom

        self.deltav = deltav
        self._set_parameters()

        if None not in (column,abundance,h2column) and column/h2column != abundance:
            raise ValueError("Cannot specify column, h2column, and abundance unless they are consistent.")

        if sum(x is None for x in (column,abundance,h2column)) >= 2:
            raise ValueError("Must specify at least two of: h2column, column, and abundance")

        if abundance:
            self._abundance = abundance
        else:
            self._abundance = column/h2column
        
        if h2column:
            self.h2column = h2column

        if column:
            self.column = column

        self.temperature = temperature
        self.tbg = tbackground

        self.debug = debug

        self.source_area = source_area

    def _set_parameters(self):

        #self.radex.cphys.cdmol = self.column
        #self.radex.cphys.tkin = self.temperature
        if u:
            self.radex.cphys.deltav = self.deltav.to(u.cm/u.s).value
        else:
            self.radex.cphys.deltav = self.deltav*1e5

        # these parameters are only used for outputs and therefore can be ignored
        self.radex.freq.fmin = 0
        self.radex.freq.fmax = 1e10

        if not hasattr(self, 'miniter'):
            self.miniter = 10
        if not hasattr(self, 'maxiter'):
            self.maxiter = 200

    _valid_colliders = ['H2','PH2','OH2','E','H','HE','H+']

    @property
    def density(self):
        d = {'H2':self.radex.cphys.density[0],
             'pH2':self.radex.cphys.density[1],
             'oH2':self.radex.cphys.density[2],
             'e':self.radex.cphys.density[3],
             'H':self.radex.cphys.density[4],
             'He':self.radex.cphys.density[5],
             'H+':self.radex.cphys.density[6]}
        if u:
            for k in d:
                d[k] = d[k] * u.cm**-3
        return d

    @density.setter
    def density(self, collider_density):

        collider_ids = {'H2': 0,
                        'PH2': 1,
                        'OH2': 2,
                        'E': 3,
                        'H': 4,
                        'HE': 5,
                        'H+': 6}

        collider_densities = defaultdict(lambda: 0)
        for k in collider_density:
            collider_densities[k.upper()] = collider_density[k]
            if k.upper() not in self._valid_colliders:
                raise ValueError('Collider %s is not one of the valid colliders: %s' %
                                 (k,self._valid_colliders))

        if 'OH2' in collider_densities:
            if not 'PH2' in collider_densities:
                raise ValueError("If o-H2 density is specified, p-H2 must also be.")
            # TODO: look up whether RADEX uses density[0] if density[1] and [2] are specified
            # (it looks like the answer is "no" based on a quick test)
            self.radex.cphys.density[0] = 0 # collider_densities['OH2'] + collider_densities['PH2']
            # PARA is [1], ORTHO is [2]
            # See lines 91, 92 of io.f
            self.radex.cphys.density[1] = collider_densities['PH2']
            self.radex.cphys.density[2] = collider_densities['OH2']
        elif 'H2' in collider_densities:
            warnings.warn("Using a default ortho-to-para ratio (which "
                          "will only affect species for which independent "
                          "ortho & para collision rates are given)")
            self.radex.cphys.density[0] = collider_densities['H2']

            T = self.temperature.value if hasattr(self.temperature,'value') else self.temperature
            if T > 0:
                opr = min(3.0,9.0*np.exp(-170.6/T))
            else:
                opr = 3.0
            fortho = opr/(1+opr)
            self.radex.cphys.density[1] = collider_densities['H2']*(1-fortho)
            self.radex.cphys.density[2] = collider_densities['H2']*(fortho)

        self.radex.cphys.density[3] = collider_densities['E']
        self.radex.cphys.density[4] = collider_densities['H']
        self.radex.cphys.density[5] = collider_densities['HE']
        self.radex.cphys.density[6] = collider_densities['H+']

        # skip H2 when computing by assuming OPR correctly distributes ortho & para
        # It's not obvious that RADEX does this correctly in readdata.f
        self.radex.cphys.totdens = self.radex.cphys.density[1:].sum()
    
    @property
    def total_density(self):
        if u:
            return self.radex.cphys.totdens * u.cm**-3
        else:
            return self.radex.cphys.totdens

    @property
    def mass_density(self):
        d = {'H2':self.radex.cphys.density[0]*0, # ignore H2, use o/p H2
             'pH2':self.radex.cphys.density[1]*2,
             'oH2':self.radex.cphys.density[2]*2,
             'e':self.radex.cphys.density[3]/1836.,
             'H':self.radex.cphys.density[4]*2,
             'He':self.radex.cphys.density[5]*4,
             'H+':self.radex.cphys.density[6]}
        if u:
            return np.sum(d.values())*constants.m_p
        else:
            return np.sum(d.values())


    @property
    def opr(self):
        return self.radex.cphys.density[1]/self.radex.cphys.density[2]

    @property
    def molpath(self):
        return "".join(self.radex.impex.molfile).strip()

    @molpath.setter
    def molpath(self, molfile):
        if "~" in molfile:
            molfile = os.path.expandpath(molfile)
        self.radex.impex.molfile[:] = ""
        utils.verify_collisionratefile(molfile)
        self.radex.impex.molfile[:len(molfile)] = molfile

    @property
    def outfile(self):
        return self.radex.impex.outfile

    @outfile.setter
    def outfile(self, outfile):
        self.radex.impex.outfile[:len(self.outfile)] = outfile

    @property
    def logfile(self):
        return self.radex.setup.logfile

    @logfile.setter
    def logfile(self, logfile):
        self.radex.setup.logfile[:len(self.logfile)] = logfile

    @property
    def datapath(self):
        return os.path.expanduser("".join(self.radex.setup.radat).strip())

    @datapath.setter
    def datapath(self, radat):
        # self.radex data path not needed if molecule given as full path
        self.radex.setup.radat[:] = ""
        self.radex.setup.radat[:len(radat)] = radat


    @property
    def escapeProbGeom(self):
        mdict = {2:'lvg',1:'sphere',3:'slab'}
        return mdict[int(self.radex.setup.method)]

    @escapeProbGeom.setter
    def escapeProbGeom(self, escapeProbGeom):
        mdict = {'lvg':2,'sphere':1,'slab':3}
        if escapeProbGeom not in mdict:
            raise ValueError("Invalid escapeProbGeom, must be one of "+",".join(mdict))
        self.radex.setup.method = mdict[escapeProbGeom]
        

    @property
    def level_population(self):
        return self.radex.collie.xpop

    @property
    def tex(self):
        if u:
            return self.radex.radi.tex * u.K
        else:
            return self.radex.radi.tex

    Tex = tex

    @property
    def tau(self):
        # taul(iline) = cddv*(xpop(n)*gstat(m)/gstat(n)-xpop(m))
        #$         /(fgaus*xt/aeinst(iline))
        return self.radex.radi.taul

    @property
    def frequency(self):
        if u:
            return self.radex.radi.spfreq * u.GHz
        else:
            return self.radex.radi.spfreq

    @property
    def temperature(self):
        if u:
            return self.radex.cphys.tkin*u.K
        else:
            return self.radex.cphys.tkin

    @temperature.setter
    def temperature(self, tkin):
        if hasattr(tkin,'to'):
            tkin = tkin.to(u.K).value
        if tkin <= 0 or tkin > 1e4:
            raise ValueError('Must have kinetic temperature > 0 and < 10^4 K')
        self.radex.cphys.tkin = tkin
        
        if not os.path.exists(self.molpath):
            raise IOError("File not found: %s" % self.molpath)
        # must re-read molecular file and re-interpolate to new temperature
        self.radex.readdata()

    @property
    def column(self):
        if u:
            return self.radex.cphys.cdmol * u.cm**-2
        else:
            return self.radex.cphys.cdmol

    @column.setter
    def column(self, col):
        if hasattr(col,'to'):
            col = col.to('cm**-2').value
        if col < 1e5 or col > 1e25:
            raise ValueError("Extremely low or extremely high column.")
        self.radex.cphys.cdmol = col

    @property
    def column_per_bin(self):
        return self.column / self.deltav

    @column_per_bin.setter
    def column_per_bin(self, cddv):
        if u:
            self.column = cddv * self.deltav.to(u.km/u.s).value
        else:
            self.column = cddv * self.deltav

    @property
    def abundance(self):
        #abund = self.column / (self.total_density * self.length)
        #if u:
        #    return abund.decompose().value
        #else:
        #    return abund
        return self._abundance

    @abundance.setter
    def abundance(self, abund):
        col = self.h2column * abund
        self._abundance = abund
        #col = abund * self.total_density * self.length
        if u:
            # need to divide the column per km/s by
            self.column = col.to(u.cm**-2).value
        else:
            self.column = col

    @property
    def h2column(self):
        return self.column / self.abundance
        #return self.total_density * self.length

    @h2column.setter
    def h2column(self, nh2, unit='cm**-2'):
        if u:
            if not hasattr(nh2,'to'):
                nh2 = nh2*u.Unit(unit)
            self.column = nh2.to(u.cm**-2) * self.abundance
        else:
            self.column = nh2 * self.abundance

    @property
    def deltav(self):
        return self._deltav

    @deltav.setter
    def deltav(self, dv):
        if u:
            if hasattr(dv,'unit'):
                self._deltav = dv.to(u.km/u.s)
            else:
                self._deltav = dv * u.km/u.s
        else:
            self._deltav = dv

    def length(self):
        """ Hard-coded, assumed length-scale """
        if u:
            return 1*u.pc
        else:
            return 1

    #@property
    #def length(self):
    #    if self.escapeProbGeom == 'lvg':
    #        return self.column_per_bin / self.total_density
    #    else:
    #        return self.h2column / self.total_density

    @property
    def debug(self):
        return self.radex.dbg.debug

    @debug.setter
    def debug(self, debug):
        self.radex.dbg.debug = debug

    @property
    def tbg(self):
        if u:
            return self.radex.cphys.tbg * u.K
        else:
            return self.radex.cphys.tbg

    @tbg.setter
    def tbg(self, tbg):
        #print("Set TBG=%f" % tbg)
        self.radex.cphys.tbg = tbg
        self.radex.backrad()

    def run_radex(self, silent=True, reuse_last=False, reload_molfile=True,
                  abs_convergence_threshold=1e-16, rel_convergence_threshold=1e-8):
        """
        Run the iterative matrix solution using a python loop

        Parameters
        ----------
        silent: bool
            Print a message when iteration is done?
        reuse_last: bool
            If this is True, the matrix iterator will start at iteration 1
            rather than iteration 0, and it will therefore repopulate the rate
            matrix based on the radiative background alone.  In principle,
            setting this to True should result in a significantly faster
            convergence; in practice, it does not.
        reload_molfile: bool
            Re-read the molecular line file?  This is needed if the collision
            rates are different and have not been updated by, e.g., changing
            the temperature (which automatically runs the `readdata` function)
        """

        if reload_molfile or self.radex.collie.ctot.sum()==0:
            self.radex.readdata()

        #self.radex.backrad()
        
        # Given the properties of *this* class, set the appropriate RADEX
        # fortran function values
        self._set_parameters()
            
        self._iter_counter = 1 if reuse_last else 0
        
        converged = np.array(False)

        last = self.level_population.copy()

        while not converged:
            if self._iter_counter >= self.maxiter:
                if not silent:
                    print("Did not converge in %i iterations, stopping." % self.maxiter)
                break

            self.radex.matrix(self._iter_counter, converged)
            level_diff = np.abs(last-self.level_population)
            frac_level_diff = level_diff/self.level_population
            if (((level_diff.sum() < abs_convergence_threshold) or
                 (frac_level_diff.sum() < rel_convergence_threshold)) and
                self._iter_counter>self.miniter):
                if not silent:
                    print("Stopped changing after %i iterations" % self._iter_counter)
                break
            last = self.level_population.copy()
            self._iter_counter += 1

        if converged and not silent:
            print("Successfully converged after %i iterations" % self._iter_counter)

        return self._iter_counter

    @property
    def quantum_number(self):
        return np.array([("".join(x)).strip() for x in grouper(self.radex.quant.qnum.T.ravel().tolist(),6)])

    @property
    def upperlevelnumber(self):
        return self.radex.imolec.iupp

    @property
    def lowerlevelnumber(self):
        return self.radex.imolec.ilow

    @property
    def upperlevelindex(self):
        return self.radex.imolec.iupp-1

    @property
    def upperlevelpop(self):
        return self.level_population[self.upperlevelindex]

    @property
    def lowerlevelindex(self):
        return self.radex.imolec.ilow-1

    @property
    def lowerlevelpop(self):
        return self.level_population[self.lowerlevelindex]

    @property
    def upperstateenergy(self):
        return self.radex.rmolec.eup

    @property
    def source_area(self):
        return self._source_area

    @source_area.setter
    def source_area(self, source_area):
        self._source_area = source_area


    @property
    def source_line_surfbrightness(self):
        return self.source_brightness - self.background_brightness

    def line_brightness_temperature(self,beamsize):
        """
        Return the line surface brightness in kelvins for a given beam area
        (Assumes the frequencies are rest frequencies)
        """
        if u:
            #return (self.line_flux * beamsize)
            # because each line has a different frequency, have to loop it
            OK_freqs = self.frequency != 0
            return u.Quantity([x.to(u.K, u.brightness_temperature(beamsize, f)).value
                               for x,f in zip(self.line_flux_density[OK_freqs],self.frequency[OK_freqs])
                               ],
                              unit=u.K)
        else:
            raise NotImplementedError("Astropy's units are required for this conversion.")

    @property
    def inds_frequencies_included(self):
        """
        The indices of the line frequencies fitted by RADEX
        (RADEX can hold up to 99999 frequencies, but usually uses ~100)
        """
        OK_freqs = self.frequency != 0
        return np.where(OK_freqs)[0]

    @property
    def source_line_brightness_temperature(self):
        """
        The surface brightness of the source assuming it is observed with a
        beam matched to its size and it has ff=1
        """
        if u:
            #return (self.line_flux * beamsize)
            # because each line has a different frequency, have to loop it
            OK_freqs = self.inds_frequencies_included
            return u.Quantity([(x*u.sr).to(u.K, u.brightness_temperature(1*u.sr, f)).value
                               for x,f in zip(self.source_line_surfbrightness[OK_freqs],
                                              self.frequency[OK_freqs])
                               ],
                              unit=u.K)
        else:
            raise NotImplementedError("Astropy's units are required for this conversion.")

    @property
    def T_B(self):
        return self.source_line_brightness_temperature

    @property
    def background_brightness(self):
        if u:
            return self.radex.radi.backi * u.erg * u.s**-1 * u.cm**-2 * u.Hz**-1 * u.sr**-1
        else:
            return self.radex.radi.backi

    @property
    def flux_density(self):
        """
        Convert the source surface brightness to a flux density by specifying
        the emitting area of the source (in steradian-equivalent units)

        This is the non-background-subtracted version
        """

        if not self.source_area:
            raise AttributeError("Need to specify a source area in order to compute the flux density")

        return self.source_brightness * self.source_area

    @property
    def line_flux_density(self):
        """
        Background-subtracted version of flux_density
        """

        if not self.source_area:
            raise AttributeError("Need to specify a source area in order to compute the flux density")

        return self.source_line_surfbrightness * self.source_area


    @property
    def source_brightness(self):

        if u:
            thc = (2 * constants.h * constants.c).cgs / u.sr
            fk = (constants.h * constants.c / constants.k_B).cgs
        else:
            thc = 3.9728913665386055e-16
            fk = 1.4387769599838154

        ftau = np.exp(-self.tau)
        xt = self._xt
        xnu = self._xnu
        bnutex = thc*xt/(np.exp(fk*xnu/self.tex)-1.0)
        toti = self.background_brightness*ftau+bnutex*(1.0-ftau)

        return toti

    @property
    def source_brightness_beta(self):
        if u:
            thc = (2 * constants.h * constants.c).cgs / u.sr
            fk = (constants.h * constants.c / constants.k_B).cgs
        else:
            thc = 3.9728913665386055e-16
            fk = 1.4387769599838154
        xt = self._xt
        xnu = self._xnu
        bnutex = thc*xt/(np.exp(fk*xnu/self.tex)-1.0)
        toti = self.background_brightness*ftau+bnutex*(1-self.beta)
        return toti

    @property
    def beta(self):
        # this will probably be faster if vectorized (translated completely
        # from fortran to python)
        return np.array([self.radex.escprob(t) for t in self.tau])

    @property
    def _xnu(self):
        """
        Line frequency in inverse cm
        """
        if u:
            return self.radex.radi.xnu * u.cm**-1
        else:
            return self.radex.radi.xnu

    @property
    def _xt(self):
        # xt = xnu**3 # cm^-1 -> cm^-3
        return self._xnu**3

    @property
    def _cddv(self):
        return self.column / self.deltav

    @property
    def statistical_weight(self):
        return self.radex.rmolec.gstat

    def get_table(self):
        T = astropy.table.Table()
        mask = self.frequency.value != 0
        T.add_column(astropy.table.Column(name='Tex',data=self.tex[mask], unit='K'))
        T.add_column(astropy.table.Column(name='tau',data=self.tau[mask], unit=''))
        T.add_column(astropy.table.Column(name='frequency',data=self.frequency[mask], unit='GHz'))
        T.add_column(astropy.table.Column(name='upperstateenergy',data=self.upperstateenergy[mask], unit='K'))
        T.add_column(astropy.table.Column(name='upperlevel',data=self.quantum_number[self.upperlevelindex[mask]], unit=''))
        T.add_column(astropy.table.Column(name='lowerlevel',data=self.quantum_number[self.lowerlevelindex[mask]], unit=''))
        T.add_column(astropy.table.Column(name='upperlevelpop',data=self.level_population[self.upperlevelindex[mask]], unit=''))
        T.add_column(astropy.table.Column(name='lowerlevelpop',data=self.level_population[self.lowerlevelindex[mask]], unit=''))
        T.add_column(astropy.table.Column(name='brightness',data=self.source_line_surfbrightness[mask]))
        T.add_column(astropy.table.Column(name='T_B',data=self.T_B)) # T_B is pre-masked
        if self.source_area:
            T.add_column(astropy.table.Column(name='flux',data=self.line_flux_density[mask]))

        return T

    def get_synthspec(self, fmin, fmax, npts=1000, **kwargs):
        """
        Generate a synthetic spectrum of the selected molecule over the
        specified frequency range.  This task is good for quick-looks but has a
        lot of overhead for generating models and should not be used for
        fitting (unless you have a conveniently small amount of data)

        Parameters
        ----------
        fmin : `~astropy.units.Quantity`
        fmax : `~astropy.units.Quantity`
            Frequency-equivalent quantity
        """
        wcs = synthspec.FrequencyArray(fmin, fmax, npts)
        S = synthspec.SyntheticSpectrum.from_RADEX(wcs, self, **kwargs)

        return S


def density_distribution(densarr, distr, moleculecolumn,
                         tauthresh=0.8, line_ids=[], **kwargs):
    """
    Compute the LVG model for a single zone with an assumed density
    *distribution* but other properties fixed.

    Parameters
    ----------
    dendarr : array
        Array of densities corresponding to the distribution function
    distr : array
        The density distribution corresponding to the density array
    moleculecolumn : quantity
        The total column density of the molecule in question.  It will be
        redistributed across the appropriate densities.  Units: cm^-2
        [this is wrong - each density will assume a too-low optical depth]
    """
    try:
        np.testing.assert_almost_equal(distr.sum(), 1)
    except AssertionError:
        raise ValueError("The distribution must be normalized.")

    if not line_ids:
        raise ValueError("Specify at least one line ID")

    meandens = {'H2': (densarr*distr).sum()}


    # Test whether the multi-slab model is reasonable by checking:
    # if the column was all at the mean density, would any lines be
    # optically thick?
    R = Radex(collider_densities=meandens, column=moleculecolumn, **kwargs)
    R.run_radex()
    if np.any(R.tau > tauthresh):
        warnings.warn(("At least one line optical depth is >{tauthresh}.  "
                       "Smoothing may be invalid.").format(tauthresh=tauthresh))

    linestrengths = []
    taus = []
    texs = []
    for dens,prob in zip(densarr,distr):
        R.density = {'H2': dens}
        try:
            R.column = moleculecolumn * prob
            R.run_radex()
        except ValueError as ex:
            if ex.args[0] == "Extremely low or extremely high column.":
                if R.column > 1e20*u.cm**-2:
                    raise ex
                else:
                    linestrengths.append(np.zeros_like(line_ids))
                    taus.append(np.zeros_like(line_ids))
                    texs.append(np.zeros_like(line_ids)+2.73)
                    continue
            else:
                raise ex

        linestrengths.append(R.source_line_brightness_temperature[line_ids])
        taus.append(R.tau[line_ids])
        texs.append(R.tex[line_ids])

    linestrengths = np.array(linestrengths)
    taus = np.array(taus)
    texs = np.array(texs)

    return R, linestrengths, linestrengths.sum(axis=0), texs, taus


def grid():
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

if any((x in sys.argv for x in ('develop','bdist'))):
    # use setuptools for develop, but nothing else
    from setuptools import setup
    Command = object
else:
    from distutils.core import setup, Command

with open('README.rst') as file:
    long_description = file.read()

#with open('CHANGES') as file:
#    long_description += file.read()

__version__ = ""
execfile('pyradex/version.py') # sets version

#import os
#if not os.path.exists('pyradex/radex/radex.so'):
#    import install_radex
#    install_radex.install_radex()

class InstallRadex(Command):
    """
    Compile the f2py radex modules needed for the python wrapper
    """

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import install_radex
        install_radex.install_radex()

class BuildRadexExecutable(Command):
    """
    Create the files radex_sphere, radex_lvg, radex_slab in the Radex/bin/
    directory
    """

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import install_radex
        install_radex.build_radex_executable()

import subprocess
import shutil
import os

class PyTest(Command):

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        if os.path.exists('build'):
            shutil.rmtree('build')
        #errno1 = subprocess.call(['py.test','--genscript=runtests.py'])
        errno2 = subprocess.call([sys.executable, 'runtests.py'])
        raise SystemExit(errno2)

setup(name='pyradex',
      version=__version__,
      description='Python-RADEX',
      long_description=long_description,
      author='Adam Ginsburg & Julia Kamenetzky',
      author_email='adam.g.ginsburg@gmail.com',
      url='http://github.com/keflavich/pyradex/',
      packages=['pyradex','pyradex.radex','pyradex.tests'],
      package_data={'pyradex.radex':['radex.so'],
                    'pyradex.tests':['data/example.out']},
      requires=['astroquery','specutils'],
      cmdclass={'test': PyTest,
                'install_radex': InstallRadex,
                'build_radex_exe': BuildRadexExecutable},
      #include_package_data=True,
      )



import numpy as np
import pyradex
from astropy.utils.console import ProgressBar

def grid_wrapper(molecule,
                 temperatures=[],
                 densities=[],
                 columns=[],
                 h2columns=[],
                 abundances=[],
                 transition_indices=[],
                 orthopararatios=[],
                 observable_parameters=['tex','source_line_surfbrightness','tau'],
                 deltav=1.0, # km/s
                 ):
    """
    Wrapper to build grids

    Parameters
    ----------
    """


    ntemp = len(temperatures)
    ndens = len(densities)
    coltype = 'h2' if h2columns else 'mol'
    columns = columns or h2columns
    ncols = len(columns) or len(h2columns)
    nabund = len(abundances)
    nopr = len(orthopararatios)

    grids = {tid:
             {par: np.empty([nopr, ncols, nabund, ntemp, ndens])
              for par in observable_parameters}
             for tid in transition_indices}

    # Just a quick first run to get things initialized
    if coltype == 'mol':
        R = pyradex.Radex(species=molecule, column=columns[0], abundance=abundances[0])
    else:
        R = pyradex.Radex(species=molecule, h2column=h2columns[0], abundance=abundances[0])
    R.deltav = deltav
    R.run_radex()

    # get the table so we can look at the frequency grid
    # table = R.get_table()

    # Target frequencies:
    # frequencies = table[np.array(transition_indices)]

    pb = ProgressBar(ntemp*ndens*ncols*nabund*nopr)

    for opri, opr in enumerate(orthopararatios):
        fortho = opr/(1+opr)
        for coli, col in enumerate(columns):
            for abundi, abund in enumerate(abundances):
                for temi,tem in enumerate(temperatures):
                    R.temperature = tem
                    for densi,dens in enumerate(densities):
                        R.density = {'oH2':dens*fortho,'pH2':dens*(1-fortho)}
                        if coltype == 'mol':
                            R.column = col
                        else:
                            R.h2column = col
                        R.abundance = abund # reset column to the appropriate value
                        R.run_radex(reuse_last=False, reload_molfile=True)

                        # type is important: MUST be tuple!
                        ind = (opri, coli, abundi, temi, densi)

                        pb.update()

                        for tid in transition_indices:
                            for par in observable_parameters:
                                val = getattr(R, par)[tid]
                                if hasattr(val,'value'):
                                    val = val.value
                                grids[tid][par][ind] = val

    return grids



import pyradex
import pytest
import os
import distutils.spawn

exepath = 'Radex/bin/radex'
#if os.path.isfile(exepath) and os.access(exepath, os.X_OK):
#    exepath = exepath
if distutils.spawn.find_executable('radex'):
    exepath = distutils.spawn.find_executable('radex')
else:
    exepath = 'radex'

def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'data')
    return os.path.join(data_dir, filename)

def test_parse_example():
    data = pyradex.parse_outfile(data_path('example.out'))
    data.pprint(show_unit=True)

@pytest.mark.skipif(exepath=='radex',reason='radex not installed')
def test_call():
    data = pyradex.pyradex(executable=exepath,species='Radex/data/hco+',minfreq=50)
    data.pprint(show_unit=True)

@pytest.mark.skipif(exepath=='radex',reason='radex not installed')
@pytest.mark.parametrize('molecule,',('co','13co','c18o','o-h2co','p-nh3',))
def test_molecules(molecule):
    if os.path.isfile('examples/%s.dat' % molecule):
        molecule = 'examples/%s' % molecule
    elif os.path.isfile('Radex/data/%s.dat' % molecule):
        molecule = 'Radex/data/%s' % molecule
    else:
        return

    data = pyradex.pyradex(executable=exepath,species=molecule,minfreq=1,maxfreq=250)
    data.pprint(show_unit=True)

def test_radex_class():
    R = pyradex.Radex(datapath='examples/',species='co',abundance=1e-4,column=1e15)
    assert hasattr(R,'radex')

def test_change_abundance():
    R = pyradex.Radex(datapath='examples/',species='co',abundance=1e-4,column=1e15)
    totdens = R.total_density
    R.abundance = 1e-6
    assert totdens == R.total_density

def test_consistent_abund():
    with pytest.raises(ValueError):
        R = pyradex.Radex(datapath='examples/',species='co',abundance=1e-4)
    with pytest.raises(ValueError):
        R = pyradex.Radex(datapath='examples/',species='co',abundance=1e-4,column=1e15,h2column=1e20)
    with pytest.raises(ValueError):
        R = pyradex.Radex(datapath='examples/',species='co',abundance=1e-4,h2column=None,column=None)
    with pytest.raises(ValueError):
        R = pyradex.Radex(datapath='examples/',species='co',abundance=None,h2column=None,column=None)

if __name__ == "__main__":
    test_call()
    test_parse_example()
    for mol in ['co','13co','c18o','o-h2co','p-nh3']:
        test_molecules(mol)
